import gzip
from multiprocessing import Process
from typing import Callable

import typer
import uvicorn
from fastapi import FastAPI, Request, Response
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.routing import APIRoute
from fastapi_pagination import add_pagination
from starlette.middleware.cors import CORSMiddleware

from ._utils import singleton
from .middlewares import process_middleware


class GzipRequest(Request):
    async def body(self) -> bytes:
        if not hasattr(self, "_body"):
            body = await super().body()
            if "gzip" in self.headers.getlist("Content-Encoding"):
                body = gzip.decompress(body)
            self._body = body
        return self._body


class GzipRoute(APIRoute):
    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def custom_route_handler(request: Request) -> Response:
            request = GzipRequest(request.scope, request.receive)
            return await original_route_handler(request)

        return custom_route_handler


@singleton
class Bali:
    def __init__(self, base_settings=None, **kwargs):
        self.base_settings = base_settings or {}
        self.kwargs = kwargs
        self._app = None

    def __getattribute__(self, attr, *args, **kwargs):
        try:
            return super().__getattribute__(attr)
        except AttributeError:
            if not self._app:
                # uvicorn entry is __call__
                if attr == '__call__':
                    self.http()
                    return getattr(self._app, attr)

                raise Exception('FastAPI App not initialized')

            return getattr(self._app, attr)

    async def __call__(self, *args, **kwargs):
        self.http()
        await self._app.__call__(*args, **kwargs)

    def _launch_http(self):
        self._app = FastAPI(**self.base_settings)
        uvicorn.run("main:app", host="0.0.0.0", port=8000, reload=True, access_log=True)

    def _launch_rpc(self):
        service = self.kwargs.get('rpc_service')
        if not service:
            raise Exception('rpc_service not provided')
        service.serve()

    def _start_all(self):
        process_http = Process(target=self._launch_http)
        process_http.start()
        process_rpc = Process(target=self._launch_rpc)
        process_rpc.start()
        process_rpc.join()
        process_http.join()

    def settings(self, **kwargs):
        self.base_settings.update(kwargs)

    def http(self):
        """load FastAPI to instance"""
        self._app = FastAPI(**self.base_settings)
        self._app.router.route_class = GzipRoute

        # routers
        for router in self.kwargs.get('routers', []):
            self._app.include_router(**router)
        # cors
        backend_cors_origins = self.kwargs.get('backend_cors_origins')
        if backend_cors_origins:
            self._app.add_middleware(GZipMiddleware)
            self._app.add_middleware(
                CORSMiddleware,
                allow_origins=[str(origin) for origin in backend_cors_origins],
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"],
            )
            self._app.middleware('http')(process_middleware)

        add_pagination(self._app)

    def launch(self, http: bool = False, rpc: bool = False):
        start_all = not any([http, rpc])
        if start_all:
            return self._start_all()

        if http:
            self._launch_http()

        if start_all or rpc:
            self._launch_rpc()

    def start(self):
        typer.run(self.launch)



from enum import Enum
from datetime import datetime, date
from decimal import Decimal

import pytz
from google.protobuf import json_format

from .dateparse import *
from .timezone import StrTzInfoType, make_aware, is_aware


class ProtobufParser(json_format._Parser):  # noqa
    def _ConvertValueMessage(self, value, message):
        """Convert a JSON representation into Value message."""
        if isinstance(value, dict):
            self._ConvertStructMessage(value, message.struct_value)
        elif isinstance(value, list):
            self._ConvertListValueMessage(value, message.list_value)
        elif value is None:
            message.null_value = 0
        elif isinstance(value, bool):
            message.bool_value = value
        elif isinstance(value, json_format.six.string_types):
            message.string_value = value
        elif isinstance(value, json_format._INT_OR_FLOAT):  # noqa
            message.number_value = value
        elif isinstance(value, date):
            message.string_value = value.isoformat()
        elif isinstance(value, datetime):
            if is_aware(value):
                value = value.astimezone(pytz.utc)
            message.string_value = value.isoformat()
        elif isinstance(value, Enum):
            message.string_value = value.name
        elif isinstance(value, Decimal):
            message.string_value = str(value)
        else:
            raise json_format.ParseError(
                'Value {0} has unexpected type {1}.'.format(value, type(value))
            )


class ProtobufPrinter(json_format._Printer):  # noqa
    def _FieldToJsonObject(self, field, value):
        """Converts field value according to Proto3 JSON Specification."""
        if field.cpp_type == json_format.descriptor.FieldDescriptor.CPPTYPE_MESSAGE:
            return self._MessageToJsonObject(value)
        elif field.cpp_type == json_format.descriptor.FieldDescriptor.CPPTYPE_ENUM:
            if self.use_integers_for_enums:
                return value
            if field.enum_type.full_name == 'google.protobuf.NullValue':
                return None
            enum_value = field.enum_type.values_by_number.get(value, None)
            if enum_value is not None:
                return enum_value.name
            else:
                if field.file.syntax == 'proto3':
                    return value
                raise json_format.SerializeToJsonError(
                    'Enum field contains an integer value '
                    'which can not mapped to an enum value.'
                )
        elif field.cpp_type == json_format.descriptor.FieldDescriptor.CPPTYPE_STRING:
            if field.type == json_format.descriptor.FieldDescriptor.TYPE_BYTES:
                # Use base64 Data encoding for bytes
                return json_format.base64.b64encode(value).decode('utf-8')
            else:
                try:
                    return datetime.fromisoformat(value)
                except ValueError:
                    return value
        elif field.cpp_type == json_format.descriptor.FieldDescriptor.CPPTYPE_BOOL:
            return bool(value)
        elif field.cpp_type in json_format._INT64_TYPES:  # noqa
            return str(value)
        elif field.cpp_type in json_format._FLOAT_TYPES:  # noqa
            if json_format.math.isinf(value):
                if value < 0.0:
                    return json_format._NEG_INFINITY  # noqa
                else:
                    return json_format._INFINITY  # noqa
            if json_format.math.isnan(value):
                return json_format._NAN  # noqa
            if self.float_format:
                return float(format(value, self.float_format))
            else:
                converted_i = int(value)
                converted_f = float(value)
                return converted_i if converted_f == converted_i else converted_f

        return value


def MessageToDict(  # noqa
        message,
        including_default_value_fields=False,
        preserving_proto_field_name=False,
        use_integers_for_enums=False,
        descriptor_pool=None,
        float_precision=None,
):
    printer = ProtobufPrinter(
        including_default_value_fields,
        preserving_proto_field_name,
        use_integers_for_enums,
        descriptor_pool,
        float_precision=float_precision
    )
    return printer._MessageToJsonObject(message)  # noqa


def ParseDict(  # noqa
    js_dict,
    message,
    ignore_unknown_fields=False,
    descriptor_pool=None,
):
    parser = ProtobufParser(ignore_unknown_fields, descriptor_pool)
    parser.ConvertMessage(js_dict, message)
    return message


def get_beginning_datetime(
    *,
    year: int,
    month: int = 1,
    day: int = 1,
    timezone: StrTzInfoType = None,
    is_dst: bool = False,
) -> datetime:
    _datetime = datetime(year, month, day)
    return make_aware(_datetime, timezone=timezone, is_dst=is_dst)



import logging
from functools import wraps

from sqla_wrapper import SQLAlchemy
from sqlalchemy.exc import OperationalError
from .models import get_base_model, AwareDateTime

# from core.config import settings

error_logger = logging.getLogger('error')


# noinspection PyPep8Naming
class DB:
    def __init__(self):
        self._session = None

    def connect(self, database_uri, **kwargs):
        kwargs.setdefault("pool_size", 5)
        kwargs.setdefault("pool_recycle", 2 * 60 * 60)

        # developers need to know when the ORM object needs to reload from the db
        kwargs.setdefault("expire_on_commit", False)
        self._session = SQLAlchemy(database_uri, **kwargs)

    def __getattribute__(self, attr, *args, **kwargs):
        try:
            return super().__getattribute__(attr)
        except AttributeError:
            if not self._session:
                raise Exception('Database session not initialized')

            # BaseModel
            if attr == 'BaseModel':
                return get_base_model(self)

            return getattr(self._session, attr)


db = DB()

MAXIMUM_RETRY_ON_DEADLOCK: int = 3


def retry_on_deadlock_decorator(func):
    lock_messages_error = ['Deadlock found', 'Lock wait timeout exceeded']

    @wraps(func)
    def wrapper(*args, **kwargs):
        attempt_count = 0
        while attempt_count < MAXIMUM_RETRY_ON_DEADLOCK:
            try:
                return func(*args, **kwargs)
            except OperationalError as e:
                # noinspection PyUnresolvedReferences
                if any(msg in e.message for msg in lock_messages_error) \
                        and attempt_count <= MAXIMUM_RETRY_ON_DEADLOCK:
                    error_logger.error(
                        'Deadlock detected. Trying sql transaction once more. Attempts count: %s' %
                        (attempt_count + 1)
                    )
                else:
                    raise
            attempt_count += 1

    return wrapper


def close_connection(func):
    def wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        finally:
            db.remove()

        return result

    return wrapper



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: helloworld.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='helloworld.proto',
  package='helloworld',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x10helloworld.proto\x12\nhelloworld\x1a\x1cgoogle/protobuf/struct.proto\"\x1c\n\x0cHelloRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x1d\n\nHelloReply\x12\x0f\n\x07message\x18\x01 \x01(\t\"\x18\n\nGetRequest\x12\n\n\x02id\x18\x01 \x01(\x05\"h\n\x0bListRequest\x12(\n\x07\x66ilters\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x0e\n\x06offset\x18\x02 \x01(\x05\x12\r\n\x05limit\x18\x03 \x01(\x05\x12\x10\n\x08ordering\x18\x04 \x03(\t\"6\n\rCreateRequest\x12%\n\x04\x64\x61ta\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\"B\n\rUpdateRequest\x12\n\n\x02id\x18\x01 \x01(\x05\x12%\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"\x1b\n\rDeleteRequest\x12\n\n\x02id\x18\x01 \x01(\x05\"5\n\x0cItemResponse\x12%\n\x04\x64\x61ta\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\"r\n\x0cListResponse\x12%\n\x04\x64\x61ta\x18\x01 \x03(\x0b\x32\x17.google.protobuf.Struct\x12\r\n\x05\x63ount\x18\x02 \x01(\x05\x12\x0e\n\x06offset\x18\x03 \x01(\x05\x12\r\n\x05limit\x18\x04 \x01(\x05\x12\r\n\x05total\x18\x05 \x01(\x05\" \n\x0eResultResponse\x12\x0e\n\x06result\x18\x01 \x01(\x08\x32\xdd\x03\n\x07Greeter\x12>\n\x08SayHello\x12\x18.helloworld.HelloRequest\x1a\x16.helloworld.HelloReply\"\x00\x12@\n\nGetGreeter\x12\x16.helloworld.GetRequest\x1a\x18.helloworld.ItemResponse\"\x00\x12\x42\n\x0bListGreeter\x12\x17.helloworld.ListRequest\x1a\x18.helloworld.ListResponse\"\x00\x12\x46\n\rCreateGreeter\x12\x19.helloworld.CreateRequest\x1a\x18.helloworld.ItemResponse\"\x00\x12=\n\x07GetItem\x12\x16.helloworld.GetRequest\x1a\x18.helloworld.ItemResponse\"\x00\x12@\n\tListItems\x12\x17.helloworld.ListRequest\x1a\x18.helloworld.ListResponse\"\x00\x12\x43\n\nCreateItem\x12\x19.helloworld.CreateRequest\x1a\x18.helloworld.ItemResponse\"\x00\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])




_HELLOREQUEST = _descriptor.Descriptor(
  name='HelloRequest',
  full_name='helloworld.HelloRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='helloworld.HelloRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=62,
  serialized_end=90,
)


_HELLOREPLY = _descriptor.Descriptor(
  name='HelloReply',
  full_name='helloworld.HelloReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='helloworld.HelloReply.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=92,
  serialized_end=121,
)


_GETREQUEST = _descriptor.Descriptor(
  name='GetRequest',
  full_name='helloworld.GetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='helloworld.GetRequest.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=123,
  serialized_end=147,
)


_LISTREQUEST = _descriptor.Descriptor(
  name='ListRequest',
  full_name='helloworld.ListRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='filters', full_name='helloworld.ListRequest.filters', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='offset', full_name='helloworld.ListRequest.offset', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='limit', full_name='helloworld.ListRequest.limit', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ordering', full_name='helloworld.ListRequest.ordering', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=149,
  serialized_end=253,
)


_CREATEREQUEST = _descriptor.Descriptor(
  name='CreateRequest',
  full_name='helloworld.CreateRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='helloworld.CreateRequest.data', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=255,
  serialized_end=309,
)


_UPDATEREQUEST = _descriptor.Descriptor(
  name='UpdateRequest',
  full_name='helloworld.UpdateRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='helloworld.UpdateRequest.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='helloworld.UpdateRequest.data', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=311,
  serialized_end=377,
)


_DELETEREQUEST = _descriptor.Descriptor(
  name='DeleteRequest',
  full_name='helloworld.DeleteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='helloworld.DeleteRequest.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=379,
  serialized_end=406,
)


_ITEMRESPONSE = _descriptor.Descriptor(
  name='ItemResponse',
  full_name='helloworld.ItemResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='helloworld.ItemResponse.data', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=408,
  serialized_end=461,
)


_LISTRESPONSE = _descriptor.Descriptor(
  name='ListResponse',
  full_name='helloworld.ListResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='helloworld.ListResponse.data', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='count', full_name='helloworld.ListResponse.count', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='offset', full_name='helloworld.ListResponse.offset', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='limit', full_name='helloworld.ListResponse.limit', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='total', full_name='helloworld.ListResponse.total', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=463,
  serialized_end=577,
)


_RESULTRESPONSE = _descriptor.Descriptor(
  name='ResultResponse',
  full_name='helloworld.ResultResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='result', full_name='helloworld.ResultResponse.result', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=579,
  serialized_end=611,
)

_LISTREQUEST.fields_by_name['filters'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_CREATEREQUEST.fields_by_name['data'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_UPDATEREQUEST.fields_by_name['data'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_ITEMRESPONSE.fields_by_name['data'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_LISTRESPONSE.fields_by_name['data'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
DESCRIPTOR.message_types_by_name['HelloRequest'] = _HELLOREQUEST
DESCRIPTOR.message_types_by_name['HelloReply'] = _HELLOREPLY
DESCRIPTOR.message_types_by_name['GetRequest'] = _GETREQUEST
DESCRIPTOR.message_types_by_name['ListRequest'] = _LISTREQUEST
DESCRIPTOR.message_types_by_name['CreateRequest'] = _CREATEREQUEST
DESCRIPTOR.message_types_by_name['UpdateRequest'] = _UPDATEREQUEST
DESCRIPTOR.message_types_by_name['DeleteRequest'] = _DELETEREQUEST
DESCRIPTOR.message_types_by_name['ItemResponse'] = _ITEMRESPONSE
DESCRIPTOR.message_types_by_name['ListResponse'] = _LISTRESPONSE
DESCRIPTOR.message_types_by_name['ResultResponse'] = _RESULTRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

HelloRequest = _reflection.GeneratedProtocolMessageType('HelloRequest', (_message.Message,), {
  'DESCRIPTOR' : _HELLOREQUEST,
  '__module__' : 'helloworld_pb2'
  # @@protoc_insertion_point(class_scope:helloworld.HelloRequest)
  })
_sym_db.RegisterMessage(HelloRequest)

HelloReply = _reflection.GeneratedProtocolMessageType('HelloReply', (_message.Message,), {
  'DESCRIPTOR' : _HELLOREPLY,
  '__module__' : 'helloworld_pb2'
  # @@protoc_insertion_point(class_scope:helloworld.HelloReply)
  })
_sym_db.RegisterMessage(HelloReply)

GetRequest = _reflection.GeneratedProtocolMessageType('GetRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETREQUEST,
  '__module__' : 'helloworld_pb2'
  # @@protoc_insertion_point(class_scope:helloworld.GetRequest)
  })
_sym_db.RegisterMessage(GetRequest)

ListRequest = _reflection.GeneratedProtocolMessageType('ListRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTREQUEST,
  '__module__' : 'helloworld_pb2'
  # @@protoc_insertion_point(class_scope:helloworld.ListRequest)
  })
_sym_db.RegisterMessage(ListRequest)

CreateRequest = _reflection.GeneratedProtocolMessageType('CreateRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEREQUEST,
  '__module__' : 'helloworld_pb2'
  # @@protoc_insertion_point(class_scope:helloworld.CreateRequest)
  })
_sym_db.RegisterMessage(CreateRequest)

UpdateRequest = _reflection.GeneratedProtocolMessageType('UpdateRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEREQUEST,
  '__module__' : 'helloworld_pb2'
  # @@protoc_insertion_point(class_scope:helloworld.UpdateRequest)
  })
_sym_db.RegisterMessage(UpdateRequest)

DeleteRequest = _reflection.GeneratedProtocolMessageType('DeleteRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEREQUEST,
  '__module__' : 'helloworld_pb2'
  # @@protoc_insertion_point(class_scope:helloworld.DeleteRequest)
  })
_sym_db.RegisterMessage(DeleteRequest)

ItemResponse = _reflection.GeneratedProtocolMessageType('ItemResponse', (_message.Message,), {
  'DESCRIPTOR' : _ITEMRESPONSE,
  '__module__' : 'helloworld_pb2'
  # @@protoc_insertion_point(class_scope:helloworld.ItemResponse)
  })
_sym_db.RegisterMessage(ItemResponse)

ListResponse = _reflection.GeneratedProtocolMessageType('ListResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTRESPONSE,
  '__module__' : 'helloworld_pb2'
  # @@protoc_insertion_point(class_scope:helloworld.ListResponse)
  })
_sym_db.RegisterMessage(ListResponse)

ResultResponse = _reflection.GeneratedProtocolMessageType('ResultResponse', (_message.Message,), {
  'DESCRIPTOR' : _RESULTRESPONSE,
  '__module__' : 'helloworld_pb2'
  # @@protoc_insertion_point(class_scope:helloworld.ResultResponse)
  })
_sym_db.RegisterMessage(ResultResponse)



_GREETER = _descriptor.ServiceDescriptor(
  name='Greeter',
  full_name='helloworld.Greeter',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=614,
  serialized_end=1091,
  methods=[
  _descriptor.MethodDescriptor(
    name='SayHello',
    full_name='helloworld.Greeter.SayHello',
    index=0,
    containing_service=None,
    input_type=_HELLOREQUEST,
    output_type=_HELLOREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetGreeter',
    full_name='helloworld.Greeter.GetGreeter',
    index=1,
    containing_service=None,
    input_type=_GETREQUEST,
    output_type=_ITEMRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListGreeter',
    full_name='helloworld.Greeter.ListGreeter',
    index=2,
    containing_service=None,
    input_type=_LISTREQUEST,
    output_type=_LISTRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateGreeter',
    full_name='helloworld.Greeter.CreateGreeter',
    index=3,
    containing_service=None,
    input_type=_CREATEREQUEST,
    output_type=_ITEMRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetItem',
    full_name='helloworld.Greeter.GetItem',
    index=4,
    containing_service=None,
    input_type=_GETREQUEST,
    output_type=_ITEMRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListItems',
    full_name='helloworld.Greeter.ListItems',
    index=5,
    containing_service=None,
    input_type=_LISTREQUEST,
    output_type=_LISTRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateItem',
    full_name='helloworld.Greeter.CreateItem',
    index=6,
    containing_service=None,
    input_type=_CREATEREQUEST,
    output_type=_ITEMRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_GREETER)

DESCRIPTOR.services_by_name['Greeter'] = _GREETER

# @@protoc_insertion_point(module_scope)



"""Generic request/response schemas
"""

from typing import List, Dict, Any

from pydantic import BaseModel, Field

__VERSION__ = 1

__all__ = [
    'GetRequest',
    'ListRequest',
    'CreateRequest',
    'UpdateRequest',
    'DeleteRequest',
    'ItemResponse',
    'ListResponse',
    'ResultResponse',
]


class GetRequest(BaseModel):
    id: int = Field(default_factory=int)


class ListRequest(BaseModel):
    filters: Dict[str, Any] = Field(default_factory=dict)
    offset: int = Field(default_factory=int)
    limit: int = Field(default_factory=int)
    ordering: List[str] = Field(default_factory=list)


class CreateRequest(BaseModel):
    data: Dict[str, Any] = Field(default_factory=dict)


class UpdateRequest(BaseModel):
    id: int = Field(default_factory=int)
    data: Dict[str, Any] = Field(default_factory=dict)


class DeleteRequest(BaseModel):
    id: int = Field(default_factory=int)


class ItemResponse(BaseModel):
    data: List[Dict[str, Any]] = Field(default_factory=list)


class ListResponse(BaseModel):
    data: List[Dict[str, Any]] = Field(default_factory=list)
    count: int = Field(default_factory=int)


class ResultResponse(BaseModel):
    result: bool = Field(default_factory=bool)



import hashlib
from functools import wraps
from urllib.parse import quote

from .cache import cache

MARKER = object()


def cache_memoize(timeout):
    def noop(*args):
        return args

    args_rewrite = noop

    def decorator(func):
        def _default_make_cache_key(*args, **kwargs):
            prefix_ = ".".join((func.__module__ or "", func.__qualname__))
            cache_key = "cache_memoize:" + ":" + prefix_ + ":".join(
                [quote(str(x)) for x in args_rewrite(args)] +
                [quote("{}={}".format(k, v)) for k, v in kwargs.items()]
            )
            return hashlib.md5(cache_key.encode('utf-8')).hexdigest()

        _make_cache_key = _default_make_cache_key

        @wraps(func)
        def inner(*args, **kwargs):
            _refresh = bool(kwargs.pop('_refresh', False))
            cache_key = _make_cache_key(*args, **kwargs)
            if _refresh:
                result = MARKER
            else:
                result = cache.get(cache_key) or MARKER

            if result is MARKER:
                result = func(*args, **kwargs)
                cache.set(cache_key, result, timeout)
            return result

        # def invalidate(*args, **kwargs):
        #     kwargs.pop("_refresh", None)
        #     cache_key = _make_cache_key(*args, **kwargs)

        return inner

    return decorator



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import helloworld_pb2 as helloworld__pb2


class GreeterStub(object):
    """The greeting service definition.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SayHello = channel.unary_unary(
                '/helloworld.Greeter/SayHello',
                request_serializer=helloworld__pb2.HelloRequest.SerializeToString,
                response_deserializer=helloworld__pb2.HelloReply.FromString,
                )
        self.GetGreeter = channel.unary_unary(
                '/helloworld.Greeter/GetGreeter',
                request_serializer=helloworld__pb2.GetRequest.SerializeToString,
                response_deserializer=helloworld__pb2.ItemResponse.FromString,
                )
        self.ListGreeter = channel.unary_unary(
                '/helloworld.Greeter/ListGreeter',
                request_serializer=helloworld__pb2.ListRequest.SerializeToString,
                response_deserializer=helloworld__pb2.ListResponse.FromString,
                )
        self.CreateGreeter = channel.unary_unary(
                '/helloworld.Greeter/CreateGreeter',
                request_serializer=helloworld__pb2.CreateRequest.SerializeToString,
                response_deserializer=helloworld__pb2.ItemResponse.FromString,
                )
        self.GetItem = channel.unary_unary(
                '/helloworld.Greeter/GetItem',
                request_serializer=helloworld__pb2.GetRequest.SerializeToString,
                response_deserializer=helloworld__pb2.ItemResponse.FromString,
                )
        self.ListItems = channel.unary_unary(
                '/helloworld.Greeter/ListItems',
                request_serializer=helloworld__pb2.ListRequest.SerializeToString,
                response_deserializer=helloworld__pb2.ListResponse.FromString,
                )
        self.CreateItem = channel.unary_unary(
                '/helloworld.Greeter/CreateItem',
                request_serializer=helloworld__pb2.CreateRequest.SerializeToString,
                response_deserializer=helloworld__pb2.ItemResponse.FromString,
                )


class GreeterServicer(object):
    """The greeting service definition.
    """

    def SayHello(self, request, context):
        """Sends a greeting
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetGreeter(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListGreeter(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateGreeter(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetItem(self, request, context):
        """item resource
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListItems(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateItem(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GreeterServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'SayHello': grpc.unary_unary_rpc_method_handler(
                    servicer.SayHello,
                    request_deserializer=helloworld__pb2.HelloRequest.FromString,
                    response_serializer=helloworld__pb2.HelloReply.SerializeToString,
            ),
            'GetGreeter': grpc.unary_unary_rpc_method_handler(
                    servicer.GetGreeter,
                    request_deserializer=helloworld__pb2.GetRequest.FromString,
                    response_serializer=helloworld__pb2.ItemResponse.SerializeToString,
            ),
            'ListGreeter': grpc.unary_unary_rpc_method_handler(
                    servicer.ListGreeter,
                    request_deserializer=helloworld__pb2.ListRequest.FromString,
                    response_serializer=helloworld__pb2.ListResponse.SerializeToString,
            ),
            'CreateGreeter': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateGreeter,
                    request_deserializer=helloworld__pb2.CreateRequest.FromString,
                    response_serializer=helloworld__pb2.ItemResponse.SerializeToString,
            ),
            'GetItem': grpc.unary_unary_rpc_method_handler(
                    servicer.GetItem,
                    request_deserializer=helloworld__pb2.GetRequest.FromString,
                    response_serializer=helloworld__pb2.ItemResponse.SerializeToString,
            ),
            'ListItems': grpc.unary_unary_rpc_method_handler(
                    servicer.ListItems,
                    request_deserializer=helloworld__pb2.ListRequest.FromString,
                    response_serializer=helloworld__pb2.ListResponse.SerializeToString,
            ),
            'CreateItem': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateItem,
                    request_deserializer=helloworld__pb2.CreateRequest.FromString,
                    response_serializer=helloworld__pb2.ItemResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'helloworld.Greeter', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Greeter(object):
    """The greeting service definition.
    """

    @staticmethod
    def SayHello(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/helloworld.Greeter/SayHello',
            helloworld__pb2.HelloRequest.SerializeToString,
            helloworld__pb2.HelloReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetGreeter(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/helloworld.Greeter/GetGreeter',
            helloworld__pb2.GetRequest.SerializeToString,
            helloworld__pb2.ItemResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListGreeter(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/helloworld.Greeter/ListGreeter',
            helloworld__pb2.ListRequest.SerializeToString,
            helloworld__pb2.ListResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateGreeter(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/helloworld.Greeter/CreateGreeter',
            helloworld__pb2.CreateRequest.SerializeToString,
            helloworld__pb2.ItemResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetItem(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/helloworld.Greeter/GetItem',
            helloworld__pb2.GetRequest.SerializeToString,
            helloworld__pb2.ItemResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListItems(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/helloworld.Greeter/ListItems',
            helloworld__pb2.ListRequest.SerializeToString,
            helloworld__pb2.ListResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateItem(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/helloworld.Greeter/CreateItem',
            helloworld__pb2.CreateRequest.SerializeToString,
            helloworld__pb2.ItemResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from typing import Any

from sqlalchemy.ext.declarative import as_declarative, declared_attr


@as_declarative()
class Base:
    id: Any
    __name__: str

    # Generate __tablename__ automatically

    @declared_attr
    def __tablename__(cls) -> str:
        return cls.__name__.lower()

    def to_dict(self):
        return {c.name: getattr(self, c.name, None) for c in self.__table__.columns}

    def dict(self):
        return self.to_dict()



from typing import Optional

from pydantic import BaseModel

from bali.db.operators import get_filters_expr
from bali.decorators import action
from bali.resource import Resource
from bali.schemas import ListRequest
from models import Item
from permissions import IsAuthenticated
from schemas import ItemModel


class QFilter(BaseModel):
    name: str


class ItemResource(Resource):

    schema = ItemModel
    filters = [
        {
            'name': Optional[str]
        },
    ]
    permission_classes = [IsAuthenticated]

    @action()
    def get(self, pk=None):
        return Item.first(id=pk)

    @action()
    def list(self, schema_in: ListRequest = None):
        return Item.query().filter(*get_filters_expr(Item, **schema_in.filters))

    @action()
    def create(self, schema_in: schema = None):
        return Item.create(**schema_in.dict())

    @action()
    def update(self, schema_in: schema = None, pk=None):
        item = Item.first(id=pk)
        for k, v in schema_in.dict():
            setattr(item, k, v)
        return item.save()

    @action()
    def delete(self, pk=None):
        item = Item.first(id=pk)
        item.delete()
        return {'result': True}

    @action(detail=False)
    def recents(self):
        """List recent 2 record"""
        items = Item.query().limit(2)
        return [self.schema(**item.dict()) for item in items]

    @action(detail=False)
    def recents_q(self, schema_in: QFilter = None):
        """List recent 2 record"""
        print('schema_in', schema_in)
        items = Item.query().limit(2)
        return [self.schema(**item.dict()) for item in items]

    @action(detail=True)
    def items_recents(self, pk=None):
        items = Item.query().filter(Item.id == pk).limit(1)
        return [self.schema(**item.dict()) for item in items]



from pydantic import BaseModel
from fastapi_pagination import paginate as default_paginate
from fastapi_pagination.ext.sqlalchemy import paginate as sqlalchemy_paginate

from .exceptions import ReturnTypeError


def paginate(sequence, params=None, is_rpc=False):
    if isinstance(sequence, BaseModel):
        raise ReturnTypeError('Paginate should return a sequence')

    # determine paginator type
    if isinstance(sequence, list):
        paginator_func = default_paginate

    else:
        paginator_func = sqlalchemy_paginate

    if params:
        paginator = paginator_func(sequence, params)
    else:
        paginator = paginator_func(sequence)

    if not is_rpc:
        return paginator

    response_data = paginator.dict()

    # convert items to dict
    items = response_data.get('items', [])
    items = [item if isinstance(item, dict) else item.dict() for item in items]

    response_data.update(
        count=response_data.get('total'),
        items=items,
        data=items,
    )

    return response_data



from bali.permissions import BasePermission


class IsAuthenticated(BasePermission):
    """
    Allows access only to authenticated users.
    """

    def has_permission(self):
        return True

    def process_auth(self):
        print(self.resource)



from datetime import datetime

import pytz
import pytest

from bali.utils import timezone


def test_get_current_timezone():
    assert timezone.get_current_timezone() == pytz.timezone(timezone.DEFAULT_TZ_INFO)


def test_get_current_timezone_name():
    assert timezone.get_current_timezone_name() == pytz.timezone(timezone.DEFAULT_TZ_INFO).tzname(None)


def test_now():
    now = timezone.now()
    assert now.utcoffset() is not None


def test_is_aware():
    now = datetime.now()
    assert not timezone.is_aware(now)


def test_is_naive():
    now = datetime.now()
    assert timezone.is_naive(now)


def test_make_aware():
    d = timezone.make_aware(datetime.now())
    assert d.tzinfo.tzname(None) == timezone.get_current_timezone().tzname(None)

    d = timezone.make_aware(datetime.now(), timezone="Asia/Shanghai")
    assert d.tzinfo.tzname(None) != timezone.get_current_timezone().tzname(None)

    d = timezone.now()
    with pytest.raises(AssertionError):
        timezone.make_aware(d)


def test_make_naive():
    d = timezone.make_naive(timezone.now())
    assert d.tzinfo is None

    d = timezone.make_naive(timezone.now(), timezone="Asia/Shanghai")
    assert d.tzinfo is None

    d = datetime.now()
    with pytest.raises(AssertionError):
        timezone.make_naive(d)



from pydantic import BaseSettings

# noinspection PyUnresolvedReferences
from .application import Bali
from .cache import cache
from .cache_memoize import cache_memoize
from .db import db, transaction
# noinspection PyUnresolvedReferences
from .routing import APIRouter

# Bind transaction to db
setattr(db, 'transaction', transaction)


class Settings(BaseSettings):
    ENABLED_RPC_LOGGING: str = False

_settings = Settings()


def initialize(settings):
    # update basic settings
    if not hasattr(settings, 'ENABLED_RPC_LOGGING'):
        settings.Config.__setattr__(settings, 'ENABLED_RPC_LOGGING', False)

    # load custom settings
    global _settings
    _settings = settings

    # initialize db connections, default enabled db
    # can be disabled by `DISABLE_DB_CONNECTION` settings
    if not getattr(settings, 'DISABLE_DB_CONNECTION', False):
        if not hasattr(settings, 'SQLALCHEMY_DATABASE_URI'):
            raise Exception('Initialized db connection without `SQLALCHEMY_DATABASE_URI` setting')

        db.connect(settings.SQLALCHEMY_DATABASE_URI)

    # initialize cache connections, default enabled db
    # can be disabled by `DISABLED_DB_CONNECTION` settings
    # cache prefix can be custom by `CACHE_PREFIX`
    if not getattr(settings, 'DISABLE_CACHE_CONNECTION', False):
        if not hasattr(settings, 'CACHE_ADDRESS'):
            raise Exception('Initialized cache connection without `CACHE_ADDRESS` setting')

        if not hasattr(settings, 'CACHE_PASSWORD'):
            raise Exception('Initialized cache connection without `CACHE_PASSWORD` setting')

        cache.connect(
            settings.CACHE_ADDRESS,
            password=settings.CACHE_PASSWORD,
            prefix=getattr(settings, 'CACHE_PREFIX', f'{settings.SERVER_NAME}_service')
        )



from pydantic import BaseModel


class BasePermission:
    """
    A base class from which all permission classes should inherit.
    """
    def __init__(self, resource):
        self.resource = resource

    def check(self):
        self.process_auth()
        return self.has_permission()

    def has_permission(self):
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        return True

    def process_auth(self):
        # self.resource.auth = BaseModel()
        pass



"""
Bali ModelResource
"""

from .resource import Resource
from ..db.operators import get_filters_expr
from ..decorators import action
from ..schemas import ListRequest

__all__ = ['ModelResource']


class ModelResource(Resource):
    """
    `ModelResource` is a fast way to implement resource layer
    for model objects
    """
    model = None
    schema = None

    @action()
    def list(self, schema_in: ListRequest = None):
        return self.model.query().filter(
            *get_filters_expr(self.model, **schema_in.filters)
        )

    @action()
    def get(self, pk=None):
        return self.model.first(id=pk)

    @action()
    def create(self, schema_in: schema = None):
        # noinspection PyUnresolvedReferences
        return self.model.create(**schema_in.dict())

    @action()
    def update(self, schema_in: schema = None, pk=None):
        item = self.model.first(id=pk)
        # noinspection PyUnresolvedReferences
        for k, v in schema_in.dict().items():
            if v is None:
                continue
            setattr(item, k, v)
        return item.save()

    @action()
    def delete(self, pk=None):
        item = self.model.first(id=pk)
        item.delete()
        return {'id': pk, 'result': True}



from typing import Optional

from pydantic import BaseModel

from bali.resources import ModelResource
from models import Item
from permissions import IsAuthenticated
from schemas import ItemModel


class QFilter(BaseModel):
    name: str


class ItemModelResource(ModelResource):
    model = Item
    schema = ItemModel
    permission_classes = [IsAuthenticated]



import re
import types
from typing import Callable

from grpc import Server


def singleton(cls):
    _instance = {}

    def inner(*args, **kwargs):
        if cls not in _instance:
            _instance[cls] = cls(*args, **kwargs)
        return _instance[cls]

    return inner


ServiceAdderNamePattern = re.compile(r"^add_.*_to_server$")


def get_service_adder(module) -> Callable[[object, Server], None]:
    namespace = vars(module)
    for i in dir(module):
        if ServiceAdderNamePattern.match(i):
            return namespace[i]



import os
from datetime import datetime
from typing import Union

import pytz

TzInfoType = Union[type(pytz.utc), pytz.tzinfo.DstTzInfo]
StrTzInfoType = Union[TzInfoType, str]
DEFAULT_TZ_INFO = "Asia/Jakarta"
NotAwareDescription = "expects an aware datetime"


def get_current_timezone() -> TzInfoType:
    tz_info = os.environ.get("TZ", DEFAULT_TZ_INFO)
    return pytz.timezone(tz_info)


def get_current_timezone_name() -> str:
    return get_current_timezone().tzname(None)


def now() -> datetime:
    return datetime.now(pytz.utc)


def is_aware(value: datetime) -> bool:
    return value.utcoffset() is not None


def is_naive(value: datetime) -> bool:
    return value.utcoffset() is None


def make_aware(
        value: datetime,
        *,
        timezone: StrTzInfoType = None,
        is_dst: bool = False,
) -> datetime:
    assert is_naive(value), "expects a naive datetime"

    if timezone is None:
        timezone = get_current_timezone()
    elif isinstance(timezone, str):
        timezone = pytz.timezone(timezone)
    else:
        pass

    return timezone.localize(value, is_dst=is_dst)


def make_naive(
        value: datetime,
        *,
        timezone: StrTzInfoType = None,
) -> datetime:
    assert is_aware(value), NotAwareDescription

    if timezone is None:
        timezone = get_current_timezone()
    elif isinstance(timezone, str):
        timezone = pytz.timezone(timezone)
    else:
        pass

    return value.astimezone(timezone).replace(tzinfo=None)


def localtime(value: datetime = None, timezone: StrTzInfoType = None):
    value, timezone = value or now(), timezone or get_current_timezone()
    if isinstance(timezone, str):
        timezone = pytz.timezone(timezone)

    assert is_aware(value), NotAwareDescription
    return value.astimezone(timezone)


def localdate(value: datetime = None, timezone: StrTzInfoType = None):
    return localtime(value, timezone).date()



import functools
import inspect
from typing import List

from fastapi import APIRouter as FastAPIRouter

from bali.core import db
from .application import GzipRoute


class APIRoute(GzipRoute):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # inject sqlalchemy scope session remove to endpoint
        if not inspect.iscoroutinefunction(self.dependant.call):
            self.dependant.call = self._inject_scoped_session_clear(self.dependant.call)

    @staticmethod
    def _inject_scoped_session_clear(endpoint: callable):
        @functools.wraps(endpoint)
        def injected_endpoint(*args, **kwargs):
            try:
                # hack Resource list action endpoint
                # ignored all query params in `kwargs`
                if endpoint.__qualname__ == 'RouterGenerator._list.<locals>.route':
                    return endpoint(*args, **{'request': kwargs.get('request')})
                return endpoint(*args, **kwargs)
            finally:
                db.remove()

        return injected_endpoint


class APIRouter(FastAPIRouter):
    def add_api_route(self, *args, **kwargs):
        kwargs.update(route_class_override=APIRoute)
        return super().add_api_route(*args, **kwargs)

    def get(self, path, *args, **kwargs):
        self.remove_api_route(path, ['GET'])
        return super().get(path, *args, **kwargs)

    def post(self, path, *args, **kwargs):
        self.remove_api_route(path, ['POST'])
        return super().post(path, *args, **kwargs)

    def patch(self, path, *args, **kwargs):
        self.remove_api_route(path, ['PATCH'])
        return super().patch(path, *args, **kwargs)

    def delete(self, path, *args, **kwargs):
        self.remove_api_route(path, ['DELETE'])
        return super().delete(path, *args, **kwargs)

    def remove_api_route(self, path: str, methods: List[str]):
        methods = set(methods)

        for r in self.routes:
            if r.path == f'{self.prefix}{path}' and r.methods == methods:
                self.routes.remove(r)




from typing import Optional

from pydantic import BaseModel
from sqlalchemy import Column, Integer, String

from bali.db import db
from bali.db.operators import get_filters_expr
from bali.decorators import action
from bali.resource import Resource
from bali.schemas import ListRequest
from permissions import IsAuthenticated

DB_URI = 'sqlite:///:memory:'

db.connect(DB_URI)


class User(db.BaseModel):
    __tablename__ = "users"
    id = Column(Integer, primary_key=True)
    username = Column(String(50), default='')
    age = Column(Integer)


# generate records
db.create_all()
lucy = User.create(**{
    'username': 'Lucy',
    'age': 13,
})
jack = User.create(**{
    'username': 'Jack',
    'age': 20,
})
groot = User.create(**{
    'username': 'Groot',
    'age': 32,
})


class UserSchema(BaseModel):
    id: int
    username: str
    age: int


class UserResource(Resource):

    schema = UserSchema
    filters = [
        {'username': str},
        {'age': Optional[str]},
    ]  # yapf: disable
    permission_classes = [IsAuthenticated]

    @action()
    def get(self, pk=None):
        return User.first(id=pk)

    @action()
    def list(self, schema_in: ListRequest = None):
        return User.query().filter(
            *get_filters_expr(User, **schema_in.filters)
        )

    @action()
    def create(self, schema_in: schema = None):
        return User.create(**schema_in.dict())

    @action()
    def update(self, schema_in: schema = None, pk=None):
        user = User.first(id=pk)
        for k, v in schema_in.dict().items():
            setattr(user, k, v)
        return user.save()

    @action()
    def delete(self, pk=None):
        User.delete(id=pk)
        return {'result': True}

    @action(detail=False)
    def recents(self):
        return User.query().all()[:2]


def test_resource_instance():
    resource = UserResource()
    assert hasattr(resource, 'as_router')
    assert resource._is_http
    assert not resource._is_rpc


def test_resource_generic_actions():
    resource = UserResource()
    get_result = resource.get(pk=1)
    assert get_result == User.first(id=1)


def test_resource_custom_actions():
    resource = UserResource()
    assert len(resource.recents()) > 0



from contextvars import ContextVar
from datetime import datetime
from typing import List, Dict

import pytz
from sqlalchemy import Column, DateTime, Boolean
from sqlalchemy.exc import IntegrityError, SQLAlchemyError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.functions import func
from sqlalchemy.types import TypeDecorator

from ..utils import timezone


class AwareDateTime(TypeDecorator):
    impl = DateTime

    @property
    def python_type(self):
        return datetime

    def process_result_value(self, value, _):
        if value is not None:
            value = timezone.make_aware(value, timezone=pytz.utc)

        return value

    def process_bind_param(self, value, _):
        if value is not None and timezone.is_aware(value):
            value = timezone.make_naive(value, timezone=pytz.utc)

        return value


context_auto_commit = ContextVar('context_auto_commit', default=True)


def get_base_model(db):
    class BaseModel(db.Model):
        __abstract__ = True

        created_time = Column(DateTime, default=datetime.utcnow)
        updated_time = Column(DateTime, default=datetime.utcnow, onupdate=datetime.utcnow)
        is_active = Column(Boolean, default=True)

        @classmethod
        def exists(cls, **attrs):
            """Returns whether an object with these attributes exists."""
            equery = cls.query().filter_by(**attrs).exists()
            return bool(db.session.query(equery).scalar())

        @classmethod
        def create(cls, **attrs):
            """Create and persist a new record for the model, and returns it."""
            return cls(**attrs).save()

        @classmethod
        def create_or_first(cls, **attrs):
            """Tries to create a new record, and if it fails
            because already exists, return the first it founds."""
            try:
                return cls.create(**attrs)
            except IntegrityError:
                db.session.rollback()
                return cls.first(**attrs)

        @classmethod
        def first(cls, **attrs):
            """Returns the first object found with these attributes."""
            return cls.query().filter_by(**attrs).first()

        @classmethod
        def first_or_error(cls, **attrs):
            """Returns the first object found with these attributes
            or raises a `ValuError` if it doesn't find one."""
            obj = cls.first(**attrs)
            if obj is None:
                raise ValueError
            return obj

        @classmethod
        def query(cls):
            return db.session.query(cls)

        def save(self):
            """Override default model's save"""
            global context_auto_commit
            db.session.add(self)
            db.session.commit() if context_auto_commit.get() else db.session.flush()
            return self

        def delete(self):
            """Override default model's delete"""
            global context_auto_commit
            db.session.delete(self)
            db.session.commit() if context_auto_commit.get() else db.session.flush()

        def _asdict(self):
            return {c.name: getattr(self, c.name, None) for c in self.__table__.columns}

        dict = to_dict = _asdict

        @classmethod
        def count(cls, **attrs) -> int:
            return db.session.query(func.count(cls.id)).filter_by(**attrs).scalar()

        @classmethod
        def get_fields(cls) -> List[str]:
            return [c.name for c in cls.__table__.columns]

        @classmethod
        def get_or_create(cls, defaults: Dict = None, **kwargs):
            instance = db.session.query(cls).filter_by(**kwargs).one_or_none()
            if instance:
                return instance, False

            instance = cls(**{**kwargs, **(defaults or {})})  # noqa
            try:
                db.session.add(instance)
                db.session.commit()
                return instance, True
            except SQLAlchemyError:
                db.session.rollback()
                instance = db.session.query(cls).filter_by(**kwargs).one()
                return instance, False

        @classmethod
        def update_or_create(cls, defaults: Dict = None, **kwargs):
            try:
                try:
                    instance = (
                        db.session.query(cls)
                        .filter_by(**kwargs)
                        .populate_existing()
                        .with_for_update()
                        .one()
                    )
                except NoResultFound:
                    instance = cls(**{**kwargs, **(defaults or {})})  # noqa
                    try:
                        db.session.add(instance)
                        db.session.commit()
                    except SQLAlchemyError:
                        db.session.rollback()
                        instance = (
                            db.session.query(cls)
                            .filter_by(**kwargs)
                            .populate_existing()
                            .with_for_update()
                            .one()
                        )
                    else:
                        return instance, True
            except SQLAlchemyError:
                db.session.rollback()
                raise
            else:
                for k, v in defaults.items():
                    setattr(instance, k, v)
                db.session.add(instance)
                db.session.commit()
                db.session.refresh(instance)
                return instance, False

    return BaseModel



from pydantic import BaseModel
from pydantic_sqlalchemy import sqlalchemy_to_pydantic

from models import Item


class ItemModel(BaseModel):
    id: int
    name: str


ItemModel = sqlalchemy_to_pydantic(Item)



import bali
from setuptools import find_packages, setup


def read(f):
    return open(f, 'r', encoding='utf-8').read()


# https://pip.pypa.io/en/latest/user_guide/#fixing-conflicting-dependencies
INSTALL_REQUIREMENTS = read('requirements.txt').splitlines()

setup(
    name='bali-core',
    version=bali.__version__,
    description='Simplify FastAPI integrate gRPC services development',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/JoshYuJump/bali',
    author='Josh.Yu',
    author_email='josh.yu_8@live.com',
    license='MIT',
    install_requires=INSTALL_REQUIREMENTS,
    packages=find_packages(),
    package_data={'bali': ['db/*.pyi']},
    include_package_data=True,
    zip_safe=False,
)



# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The Python implementation of the GRPC helloworld.Greeter server."""

import logging
from concurrent import futures

import grpc

import helloworld_pb2
import helloworld_pb2 as pb2
import helloworld_pb2_grpc
from bali.interceptors import ProcessInterceptor
from bali.mixins import ServiceMixin
from resources import GreeterResource, ItemResource


class GrpcServer(helloworld_pb2_grpc.GreeterServicer, ServiceMixin):
    def SayHello(self, request, context):
        print('Greeter.SayHello')
        return helloworld_pb2.HelloReply(message='Hello, %s!' % request.name)

    def GetGreeter(self, request, context):
        print('Greeter.GetGreeter')
        return GreeterResource(request, context, pb2.ItemResponse).get()

    def ListGreeter(self, request, context):
        print('Greeter.ListGreeter')
        return GreeterResource(request, context, pb2.ListResponse).list()

    def CreateGreeter(self, request, context):
        print('Greeter.CreateGreeter')
        return GreeterResource(request, context, pb2.ItemResponse).create()

    def GetItem(self, request, context):
        return ItemResource(request, context, pb2.ItemResponse).get()

    def ListItems(self, request, context):
        return ItemResource(request, context, pb2.ListResponse).list()


def serve():
    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=10),
        interceptors=[ProcessInterceptor()],
    )
    helloworld_pb2_grpc.add_GreeterServicer_to_server(GrpcServer(), server)
    server.add_insecure_port('[::]:50051')
    server.start()
    print("gRPC Service Hello world started")
    server.wait_for_termination()


if __name__ == '__main__':
    logging.basicConfig()
    serve()



# noinspection PyUnresolvedReferences
import config
import grpc_server
from bali.core import Bali
from v1.app import router
from fastapi_pagination import LimitOffsetPage, add_pagination, paginate


app = Bali(
    base_settings=None,
    routers=[{
        'router': router,
        'prefix': '/v1',
    }],
    backend_cors_origins=['http://127.0.0.1'],
    rpc_service=grpc_server,
)
app.settings(title='Bali Example')

if __name__ == "__main__":
    app.start()



import logging
from .core import db


# noinspection PyMethodMayBeStatic
class ServiceMixin:
    def __init__(self):
        self.db = db

    def setup(self):
        pass

    def teardown(self):
        # noinspection PyBroadException
        try:
            db.remove()
        except Exception:
            pass

    def __getattribute__(self, attr, *args, **kwargs):
        if attr in ['setup', 'teardown']:
            return super().__getattribute__(attr)

        self.setup()
        result = super().__getattribute__(attr)
        self.teardown()
        return result
